﻿using System;
using System.ComponentModel;
using System.Diagnostics;				// Process
using System.Drawing;
using System.IO;						// File
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace AsyncTest {
	public partial class Form1 : Form {
		private Stopwatch stopwatch;
		private const int N=5;
		private int sleep;
		private int maxListboxLineLength;
		private System.Windows.Forms.Timer formsTimer;
		private System.Threading.Timer threadTimer;
		private System.Timers.Timer timersTimer;
		private int formsTimerCount;
		private int threadTimerCount;
		private int timersTimerCount;
		private int mainThreadID;
		private byte[] fileBytes=new byte[1024];
		private FileStream fileStream;
		private BackgroundWorker BGW;

		public Form1() {
			InitializeComponent();
			Width=SystemInformation.PrimaryMonitorSize.Width-200;
			Height=SystemInformation.PrimaryMonitorSize.Height-100;
			stopwatch=new Stopwatch();
			stopwatch.Start();
			mainThreadID=Thread.CurrentThread.ManagedThreadId;
			log("mainThreadID="+mainThreadID.ToString("X2")+"  (Thread IDs are shown in hex)");
		}

		private void log(string s) {
			string threadIDString=" [GUI] ";
			Thread thread=Thread.CurrentThread;
			int threadID=thread.ManagedThreadId;
			Color color=Color.MediumSeaGreen;
			if (threadID!=mainThreadID) {
				char isThreadpoolThread=' ';
				color=Color.Red;
				if (thread.IsThreadPoolThread) {
					isThreadpoolThread='P';
					color=Color.Black;
				}
				threadIDString=" ["+isThreadpoolThread+threadID.ToString("X2")+"] ";
			}
			if (s.Length!=0) s=stopwatch.Elapsed.TotalSeconds.ToString("N3")+threadIDString+s;
			Console.WriteLine(s);
			log1(new ListboxItem(s, color));
		}

		private void log1(ListboxItem item) {
			if (lb.InvokeRequired) {
				lb.Invoke(new Action<ListboxItem>(log1), new object[] { item });
			} else {
				lb.Items.Add(item);
				lb.TopIndex=lb.Items.Count-1;
				int len=item.Text.Length;
			}
		}

		private void lb_DrawItem(object sender, DrawItemEventArgs e) {
			Graphics g=e.Graphics;
			ListboxItem item=(ListboxItem)lb.Items[e.Index];
			// tell the Listbox how wide it should be to not need a horizontal scrollbar
			// (using a monospaced font, string.Length is good enough to decide on growth)
			int len=item.Text.Length;
			if (maxListboxLineLength<len) {
				maxListboxLineLength=len;
				lb.HorizontalExtent=20+(int)g.MeasureString(item.Text, lb.Font).Width;
			}
			// draw the string
			g.DrawString(item.Text, lb.Font, item.Brush, e.Bounds);
		}

		private void btnClear_Click(object sender, EventArgs e) {
			lb.Items.Clear();
			maxListboxLineLength=0;
			lb.HorizontalExtent=0;
		}

		private void btnStart_Click(object sender, EventArgs e) {
			sleep=cbSleep.Checked?500:0;
			log("");
			log("btnStart_Click");
			doThreadPool();
			doThread();
			doBackgroundWorker();
			doFile();
			doProcess();
			doTimers();
			log("btnStart_Click done");
			log("");
		}

		private void doThreadPool() {
			log("Queueing some ThreadPool jobs");
			for (int i=0; i<N; i++) ThreadPool.QueueUserWorkItem(threadPool_Callback);
		}

		private void threadPool_Callback(object state) {
			log("threadPool_Callback");
			if (sleep!=0) Thread.Sleep(sleep);
		}

		private void doThread() {
			log("Launching some explicit threads");
			for (int i=0; i<N; i++) {
				Thread thread=new Thread(thread_Code);
				thread.Start();
			}
		}

		private void thread_Code(object state) {
			log("thread_Code");
			if (sleep!=0) Thread.Sleep(sleep);
		}

		private void doBackgroundWorker() {
			log("Launching some BackgroundWorkers");
			for (int i=0; i<N; i++) {
				BGW=new BackgroundWorker();
				BGW.DoWork+=new DoWorkEventHandler(backgroundWorker_DoWork);
				BGW.WorkerReportsProgress=true;
				BGW.ProgressChanged+=new ProgressChangedEventHandler(backgroundWorker_ProgressChanged);
				BGW.RunWorkerCompleted+=new RunWorkerCompletedEventHandler(backgroundWorker_RunWorkerCompleted);
				BGW.RunWorkerAsync();
			}
		}

		void backgroundWorker_DoWork(object sender, DoWorkEventArgs e) {
			log("backgroundWorker_DoWork");
			BGW.ReportProgress(50);
			if (sleep!=0) Thread.Sleep(sleep);
		}

		void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e) {
			log("backgroundWorker_ProgressChanged: "+e.ProgressPercentage);
		}

		void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
			log("backgroundWorker_RunWorkerCompleted");
		}

		private void doFile() {
			string filename=Application.ExecutablePath;
			log("Starting async file read: "+filename);
			fileStream=File.OpenRead(filename);
			fileStream.BeginRead(fileBytes, 0, fileBytes.Length, readFile_Callback, null);
		}

		private void readFile_Callback(IAsyncResult result) {
			fileStream.Close();
			log("readFile_Callback: completed="+result.IsCompleted);
			Thread.Sleep(1000); // don't mix with doProcess() output
			for(int j=0; j<256; j+=32) {
				string s="readFile_Callback: ["+j.ToString("X4")+"]=";
				for (int i=0; i<32; i++) s+=" "+fileBytes[i+j].ToString("X2");
				log(s);
			}
		}

		private void doProcess() {
			log("Starting DIR command in external process");
			Process process=new Process();
			process.StartInfo.FileName="cmd.exe";
			process.StartInfo.Arguments="/C dir *.*";
			process.StartInfo.UseShellExecute=false;
			process.StartInfo.CreateNoWindow=true;
			process.StartInfo.RedirectStandardOutput=true;
			process.Start();
			process.OutputDataReceived+=new DataReceivedEventHandler(process_OutputDataReceived);
			process.BeginOutputReadLine();
		}

		private void process_OutputDataReceived(object sender, DataReceivedEventArgs e) {
			log("process_OutputDataReceived: "+e.Data);
			if (sleep!=0) Thread.Sleep(sleep);
		}

		private void doTimers() {
			log("Starting formsTimer");
			formsTimer=new System.Windows.Forms.Timer();
			formsTimer.Interval=1000;
			formsTimer.Tick+=new EventHandler(formsTimer_Tick);
			formsTimerCount=0;
			formsTimer.Start();
			log("Starting threadTimer");
			threadTimerCount=0;
			threadTimer=new System.Threading.Timer(threadTimer_Callback, null, 500, 1000);
			log("Starting timersTimer");
			timersTimerCount=0;
			timersTimer=new System.Timers.Timer(1000);
			timersTimer.Start();
			timersTimer.Elapsed+=new System.Timers.ElapsedEventHandler(timersTimer_Elapsed);
		}

		private void formsTimer_Tick(object sender, EventArgs e) {
			formsTimerCount++;
			log("formsTimer_Tick: "+formsTimerCount);
			if (formsTimerCount>=N) formsTimer.Stop();
			if (sleep!=0) Thread.Sleep(sleep);
		}

		private void threadTimer_Callback(object state) {
			threadTimerCount++;
			log("threadTimer_Callback: "+threadTimerCount);
			if (threadTimerCount>=N) threadTimer.Dispose();
			if (sleep!=0) Thread.Sleep(sleep);
		}

		private void timersTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e) {
			timersTimerCount++;
			log("timersTimer_Elapsed: "+timersTimerCount);
			if (timersTimerCount>=N) timersTimer.Stop();
			if (sleep!=0) Thread.Sleep(sleep);
		}
	}

	public class ListboxItem {
		public string Text;
		public Brush Brush;
		public ListboxItem(string text, Color color) {
			Text=text;
			Brush=new SolidBrush(color);
		}
	}
}
